@@ -1,11 +1,8 @@
-from __future__ import unicode_literals
-
 from django.contrib.contenttypes.fields import GenericForeignKey
 from django.contrib.contenttypes.models import ContentType
 from django.db import IntegrityError, models, router, transaction
-from django.utils.encoding import python_2_unicode_compatible
 from django.utils.text import slugify
-from django.utils.translation import ugettext, ugettext_lazy as _
+from django.utils.translation import gettext, gettext_lazy as _
 
 try:
     from unidecode import unidecode
@@ -15,10 +12,9 @@ except ImportError:
         return tag
 
 
-@python_2_unicode_compatible
 class TagBase(models.Model):
-    name = models.CharField(verbose_name=_("Name"), unique=True, max_length=100)
-    slug = models.SlugField(verbose_name=_("Slug"), unique=True, max_length=100)
+    name = models.CharField(verbose_name=_("name"), unique=True, max_length=100)
+    slug = models.SlugField(verbose_name=_("slug"), unique=True, max_length=100)
 
     def __str__(self):
         return self.name
@@ -42,31 +38,31 @@ class TagBase(models.Model):
             # with a multi-master setup, theoretically we could try to
             # write and rollback on different DBs
             kwargs["using"] = using
-            # Be oportunistic and try to save the tag, this should work for
+            # Be opportunistic and try to save the tag, this should work for
             # most cases ;)
             try:
                 with transaction.atomic(using=using):
-                    res = super(TagBase, self).save(*args, **kwargs)
+                    res = super().save(*args, **kwargs)
                 return res
             except IntegrityError:
                 pass
             # Now try to find existing slugs with similar names
             slugs = set(
-                self.__class__._default_manager.filter(
-                    slug__startswith=self.slug
-                ).values_list("slug", flat=True)
+                type(self)
+                ._default_manager.filter(slug__startswith=self.slug)
+                .values_list("slug", flat=True)
             )
             i = 1
             while True:
                 slug = self.slugify(self.name, i)
                 if slug not in slugs:
                     self.slug = slug
-                    # We purposely ignore concurrecny issues here for now.
+                    # We purposely ignore concurrency issues here for now.
                     # (That is, till we found a nice solution...)
-                    return super(TagBase, self).save(*args, **kwargs)
+                    return super().save(*args, **kwargs)
                 i += 1
         else:
-            return super(TagBase, self).save(*args, **kwargs)
+            return super().save(*args, **kwargs)
 
     def slugify(self, tag, i=None):
         slug = slugify(unidecode(tag))
@@ -77,15 +73,14 @@ class TagBase(models.Model):
 
 class Tag(TagBase):
     class Meta:
-        verbose_name = _("Tag")
-        verbose_name_plural = _("Tags")
+        verbose_name = _("tag")
+        verbose_name_plural = _("tags")
         app_label = "taggit"
 
 
-@python_2_unicode_compatible
 class ItemBase(models.Model):
     def __str__(self):
-        return ugettext("%(object)s tagged with %(tag)s") % {
+        return gettext("%(object)s tagged with %(tag)s") % {
             "object": self.content_object,
             "tag": self.tag,
         }
@@ -107,15 +102,6 @@ class ItemBase(models.Model):
     def lookup_kwargs(cls, instance):
         return {"content_object": instance}
 
-
-class TaggedItemBase(ItemBase):
-    tag = models.ForeignKey(
-        Tag, related_name="%(app_label)s_%(class)s_items", on_delete=models.CASCADE
-    )
-
-    class Meta:
-        abstract = True
-
     @classmethod
     def tags_for(cls, model, instance=None, **extra_filters):
         kwargs = extra_filters or {}
@@ -126,11 +112,20 @@ class TaggedItemBase(ItemBase):
         return cls.tag_model().objects.filter(**kwargs).distinct()
 
 
+class TaggedItemBase(ItemBase):
+    tag = models.ForeignKey(
+        Tag, related_name="%(app_label)s_%(class)s_items", on_delete=models.CASCADE
+    )
+
+    class Meta:
+        abstract = True
+
+
 class CommonGenericTaggedItemBase(ItemBase):
     content_type = models.ForeignKey(
         ContentType,
         on_delete=models.CASCADE,
-        verbose_name=_("Content type"),
+        verbose_name=_("content type"),
         related_name="%(app_label)s_%(class)s_tagged_items",
     )
     content_object = GenericForeignKey()
@@ -147,24 +142,28 @@ class CommonGenericTaggedItemBase(ItemBase):
 
     @classmethod
     def tags_for(cls, model, instance=None, **extra_filters):
-        ct = ContentType.objects.get_for_model(model)
-        kwargs = {"%s__content_type" % cls.tag_relname(): ct}
+        tag_relname = cls.tag_relname()
+        model = model._meta.concrete_model
+        kwargs = {
+            "%s__content_type__app_label" % tag_relname: model._meta.app_label,
+            "%s__content_type__model" % tag_relname: model._meta.model_name,
+        }
         if instance is not None:
-            kwargs["%s__object_id" % cls.tag_relname()] = instance.pk
+            kwargs["%s__object_id" % tag_relname] = instance.pk
         if extra_filters:
             kwargs.update(extra_filters)
         return cls.tag_model().objects.filter(**kwargs).distinct()
 
 
 class GenericTaggedItemBase(CommonGenericTaggedItemBase):
-    object_id = models.IntegerField(verbose_name=_("Object id"), db_index=True)
+    object_id = models.IntegerField(verbose_name=_("object ID"), db_index=True)
 
     class Meta:
         abstract = True
 
 
 class GenericUUIDTaggedItemBase(CommonGenericTaggedItemBase):
-    object_id = models.UUIDField(verbose_name=_("Object id"), db_index=True)
+    object_id = models.UUIDField(verbose_name=_("object ID"), db_index=True)
 
     class Meta:
         abstract = True
@@ -172,7 +171,8 @@ class GenericUUIDTaggedItemBase(CommonGenericTaggedItemBase):
 
 class TaggedItem(GenericTaggedItemBase, TaggedItemBase):
     class Meta:
-        verbose_name = _("Tagged Item")
-        verbose_name_plural = _("Tagged Items")
+        verbose_name = _("tagged item")
+        verbose_name_plural = _("tagged items")
         app_label = "taggit"
         index_together = [["content_type", "object_id"]]
+        unique_together = [["content_type", "object_id", "tag"]]
