@@ -1,17 +1,14 @@
-from __future__ import absolute_import, unicode_literals
+from unittest import mock
 
-import unittest
-
-import mock
 from django.contrib.contenttypes.models import ContentType
 from django.core import serializers
 from django.core.exceptions import ValidationError
-from django.db import connection, models
-from django.test import RequestFactory, TestCase
+from django.db import IntegrityError, connection, models
+from django.test import RequestFactory, SimpleTestCase, TestCase
 from django.test.utils import override_settings
-from django.utils.encoding import force_text
 
 from .forms import (
+    BlankTagForm,
     CustomPKFoodForm,
     DirectCustomPKFoodForm,
     DirectFoodForm,
@@ -31,9 +28,13 @@ from .models import (
     DirectFood,
     DirectHousePet,
     DirectPet,
+    DirectTrackedFood,
+    DirectTrackedHousePet,
+    DirectTrackedPet,
     Food,
     HousePet,
     Movie,
+    Name,
     OfficialFood,
     OfficialHousePet,
     OfficialPet,
@@ -41,11 +42,17 @@ from .models import (
     OfficialThroughModel,
     Pet,
     Photo,
+    ProxyPhoto,
     TaggedCustomPK,
     TaggedCustomPKFood,
     TaggedFood,
+    TaggedTrackedFood,
+    TrackedTag,
     UUIDFood,
+    UUIDHousePet,
+    UUIDPet,
     UUIDTag,
+    UUIDTaggedItem,
 )
 
 from taggit.managers import TaggableManager, _TaggableManager
@@ -114,6 +121,26 @@ class TagModelTestCase(BaseTaggingTestCase):
         self.assertIs(low < high, False)
 
 
+class CustomTagCreationTestCase(TestCase):
+    def test_model_manager_add(self):
+        apple = OfficialFood.objects.create(name="apple")
+
+        # let's add two official tags
+        apple.tags.add(
+            "foo", "bar", tag_kwargs={"official": True},
+        )
+
+        # and two unofficial ones
+        apple.tags.add(
+            "baz", "wow", tag_kwargs={"official": False},
+        )
+
+        # We should end up with 4 tags
+        self.assertEquals(apple.tags.count(), 4)
+        self.assertEquals(apple.tags.filter(official=True).count(), 2)
+        self.assertEquals(apple.tags.filter(official=False).count(), 2)
+
+
 class TagModelDirectTestCase(TagModelTestCase):
     food_model = DirectFood
     tag_model = Tag
@@ -146,6 +173,10 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
     taggeditem_model = TaggedItem
     tag_model = Tag
 
+    def setUp(self):
+        super().setUp()
+        ContentType.objects.clear_cache()
+
     def test_add_tag(self):
         apple = self.food_model.objects.create(name="apple")
         self.assertEqual(list(apple.tags.all()), [])
@@ -501,7 +532,7 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         food_instance = self.food_model()
         msg = (
             "%s objects need to have a primary key value before you can access "
-            "their tags." % self.food_model().__class__.__name__
+            "their tags." % type(self.food_model()).__name__
         )
         with self.assertRaisesMessage(ValueError, msg):
             food_instance.tags.all()
@@ -558,7 +589,7 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         model_name = self.pet_model.__name__
         self.assertQuerysetEqual(
             pks,
-            ["<{0}: kitty>".format(model_name), "<{0}: cat>".format(model_name)],
+            ["<{}: kitty>".format(model_name), "<{}: cat>".format(model_name)],
             ordered=False,
         )
 
@@ -575,7 +606,7 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         model_name = self.food_model.__name__
         self.assertQuerysetEqual(
             pks,
-            ["<{0}: pear>".format(model_name), "<{0}: guava>".format(model_name)],
+            ["<{}: pear>".format(model_name), "<{}: guava>".format(model_name)],
             ordered=False,
         )
 
@@ -609,13 +640,12 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
             self.tag_model.objects.filter(**lookup_kwargs), ["scary"]
         )
 
-    def test_taggeditem_unicode(self):
+    def test_taggeditem_str(self):
         apple = self.food_model.objects.create(name="apple")
         apple.tags.add("juicy")
 
         self.assertEqual(
-            force_text(self.taggeditem_model.objects.all()[0]),
-            "apple tagged with juicy",
+            str(self.taggeditem_model.objects.first()), "apple tagged with juicy"
         )
 
     def test_abstract_subclasses(self):
@@ -627,6 +657,17 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         m.tags.add("hd")
         self.assert_tags_equal(m.tags.all(), ["hd"])
 
+    def test_proxy_subclasses(self):
+        p = Photo.objects.create()
+        proxy_p = ProxyPhoto.objects.create()
+        p.tags.add("outdoors", "pretty")
+        self.assert_tags_equal(p.tags.all(), ["outdoors", "pretty"])
+        self.assert_tags_equal(proxy_p.tags.all(), [])
+
+        proxy_p.tags.add("hd")
+        self.assert_tags_equal(proxy_p.tags.all(), ["hd"])
+        self.assert_tags_equal(p.tags.all(), ["outdoors", "pretty"])
+
     def test_field_api(self):
         # Check if tag field, which simulates m2m, has django-like api.
         field = self.food_model._meta.get_field("tags")
@@ -639,13 +680,13 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         apple = self.food_model.objects.create(name="apple")
         apple.tags.add("green")
         apple.tags.add("red")
-        self.assertEqual(list(apple.tags.names()), ["green", "red"])
+        self.assertEqual(sorted(list(apple.tags.names())), ["green", "red"])
 
     def test_slugs_method(self):
         apple = self.food_model.objects.create(name="apple")
         apple.tags.add("green and juicy")
         apple.tags.add("red")
-        self.assertEqual(list(apple.tags.slugs()), ["green-and-juicy", "red"])
+        self.assertEqual(sorted(list(apple.tags.slugs())), ["green-and-juicy", "red"])
 
     def test_serializes(self):
         apple = self.food_model.objects.create(name="apple")
@@ -703,6 +744,19 @@ class TaggableManagerTestCase(BaseTaggingTestCase):
         tag_names = sorted(orange.tags.names())
         self.assertEqual(tag_names, ["Spain", "Valencia"])
 
+    def test_tag_uniqueness(self):
+        apple = self.food_model.objects.create(name="apple")
+        tag = self.tag_model.objects.create(name="juice", slug="juicy")
+        self.taggeditem_model.objects.create(tag=tag, content_object=apple)
+        with self.assertRaises(IntegrityError):
+            self.taggeditem_model.objects.create(tag=tag, content_object=apple)
+
+    def test_most_common_lazy(self):
+        with self.assertNumQueries(0):
+            qs = self.food_model.tags.most_common()
+        with self.assertNumQueries(1):
+            list(qs)
+
 
 class TaggableManagerDirectTestCase(TaggableManagerTestCase):
     food_model = DirectFood
@@ -711,6 +765,14 @@ class TaggableManagerDirectTestCase(TaggableManagerTestCase):
     taggeditem_model = TaggedFood
 
 
+class TaggableManagerDirectTrackedTestCase(TaggableManagerTestCase):
+    food_model = DirectTrackedFood
+    pet_model = DirectTrackedPet
+    housepet_model = DirectTrackedHousePet
+    taggeditem_model = TaggedTrackedFood
+    tag_model = TrackedTag
+
+
 class TaggableManagerDirectCustomPKTestCase(TaggableManagerTestCase):
     food_model = DirectCustomPKFood
     pet_model = DirectCustomPKPet
@@ -735,6 +797,19 @@ class TaggableManagerCustomPKTestCase(TaggableManagerTestCase):
         pass
 
 
+class TaggableManagerUUIDTestCase(TaggableManagerTestCase):
+    food_model = UUIDFood
+    pet_model = UUIDPet
+    housepet_model = UUIDHousePet
+    taggeditem_model = UUIDTaggedItem
+    tag_model = UUIDTag
+
+    def test_require_pk(self):
+        # With a UUIDField pk, pk is never None. So taggit has no way to tell
+        # if the instance is saved or not.
+        pass
+
+
 class TaggableManagerOfficialTestCase(TaggableManagerTestCase):
     food_model = OfficialFood
     pet_model = OfficialPet
@@ -802,10 +877,10 @@ class TaggableManagerInitializationTestCase(TaggableManagerTestCase):
     custom_manager_model = CustomManager
 
     def test_default_manager(self):
-        self.assertEqual(self.food_model.tags.__class__, _TaggableManager)
+        self.assertIs(type(self.food_model.tags), _TaggableManager)
 
     def test_custom_manager(self):
-        self.assertEqual(self.custom_manager_model.tags.__class__, CustomManager.Foo)
+        self.assertIs(type(self.custom_manager_model.tags), CustomManager.Foo)
 
 
 class TaggableFormTestCase(BaseTaggingTestCase):
@@ -885,6 +960,67 @@ class TaggableFormTestCase(BaseTaggingTestCase):
         ff = tm.formfield()
         self.assertRaises(ValidationError, ff.clean, "")
 
+    def test_form_changed_data(self):
+        # new food, blank tag
+        pear = self.food_model()
+        request = {"name": "pear", "tags": ""}
+        fff = self.form_class(request, instance=pear)
+        self.assertFalse(fff.is_valid())
+
+        pear = self.food_model()
+        request = {"name": "pear", "tags": "sweat"}
+        fff = self.form_class(request, instance=pear)
+        self.assertTrue(fff.is_valid())
+        self.assertIn("tags", fff.changed_data)
+        self.assertIn("name", fff.changed_data)
+        fff.save()
+
+        request = {"name": "pear", "tags": "yellow"}
+        fff = self.form_class(request, instance=pear)
+        self.assertTrue(fff.is_valid())
+        self.assertIn("tags", fff.changed_data)
+        self.assertNotIn("name", fff.changed_data)
+        fff.save()
+
+        # same object nothing changed
+        fff = self.form_class(request, instance=pear)
+        self.assertTrue(fff.is_valid())
+        self.assertFalse(fff.changed_data)
+
+        # delete tag
+        request = {"name": "pear", "tags": ""}
+        fff = self.form_class(request, instance=pear)
+        self.assertFalse(fff.is_valid())  # tag not blank
+
+        # change name, tags are the same
+        request = {"name": "apple", "tags": "yellow"}
+        fff = self.form_class(request, instance=pear)
+        self.assertTrue(fff.is_valid())
+        self.assertIn("name", fff.changed_data)
+        self.assertNotIn("tags", fff.changed_data)
+        fff.save()
+
+        # tags changed
+        apple = self.food_model.objects.get(name="apple")
+        request = {"name": "apple", "tags": "yellow, delicious"}
+        fff = self.form_class(request, instance=apple)
+        self.assertTrue(fff.is_valid())
+        self.assertNotIn("name", fff.changed_data)
+        self.assertIn("tags", fff.changed_data)
+        fff.save()
+
+        # only tags order changed
+        apple = self.food_model.objects.get(name="apple")
+        request = {"name": "apple", "tags": "delicious, yellow"}
+        fff = self.form_class(request, instance=apple)
+        self.assertTrue(fff.is_valid())
+        self.assertFalse(fff.changed_data)
+
+        # and nothing changed
+        fff = self.form_class(request, instance=apple)
+        self.assertTrue(fff.is_valid())
+        self.assertFalse(fff.changed_data)
+
 
 class TaggableFormDirectTestCase(TaggableFormTestCase):
     form_class = DirectFoodForm
@@ -906,7 +1042,25 @@ class TaggableFormOfficialTestCase(TaggableFormTestCase):
     food_model = OfficialFood
 
 
-class TagStringParseTestCase(unittest.TestCase):
+class BlankFormTestCase(SimpleTestCase):
+    form_class = BlankTagForm
+
+    def test_early_access_to_changed_data_with(self):
+        # For example in ModelForm.validate_unique().
+        request = {"name": "pear"}
+        form = self.form_class(request)
+        self.assertNotIn("tags", form.changed_data)
+
+        request = {"name": "pear", "tags": ""}
+        form = self.form_class(request)
+        self.assertNotIn("tags", form.changed_data)
+
+        request = {"name": "pear", "tags": "tag1"}
+        form = self.form_class(request)
+        self.assertIn("tags", form.changed_data)
+
+
+class TagStringParseTestCase(SimpleTestCase):
     """
     Ported from Jonathan Buchanan's `django-tagging
     <http://django-tagging.googlecode.com/>`_
@@ -1012,7 +1166,7 @@ class TagStringParseTestCase(unittest.TestCase):
         self.assertEqual(edit_string_for_tags([a, b]), "Cued Speech, transliterator")
 
 
-class DeconstructTestCase(unittest.TestCase):
+class DeconstructTestCase(SimpleTestCase):
     def test_deconstruct_kwargs_kept(self):
         instance = TaggableManager(through=OfficialThroughModel, to="dummy.To")
         name, path, args, kwargs = instance.deconstruct()
@@ -1044,7 +1198,7 @@ class TagListViewTests(TestCase):
     model = Food
 
     def setUp(self):
-        super(TagListViewTests, self).setUp()
+        super().setUp()
         self.factory = RequestFactory()
         self.slug = "green"
         self.apple = self.model.objects.create(name="apple")
@@ -1068,3 +1222,17 @@ class TagListViewTests(TestCase):
         self.assertEqual(response.status_code, 200)
         self.assertIn(self.apple, response.context_data["object_list"])
         self.assertNotIn(self.strawberry, response.context_data["object_list"])
+
+
+class RelatedNameTests(TestCase):
+    def test_default_related_name(self):
+        food = Food.objects.create(name="apple")
+        food.tags.add("green")
+        tag = Tag.objects.get(food=food.pk)
+        self.assertEqual(tag.name, "green")
+
+    def test_custom_related_name(self):
+        name = Name.objects.create()
+        name.tags.add("foo")
+        tag = Tag.objects.get(a_unique_related_name=name.pk)
+        self.assertEqual(tag.name, "foo")
